import datetime
import time

from autowsgr.constants import literals
from autowsgr.constants.custom_exceptions import ShipNotFoundErr
from autowsgr.fight.normal_fight import NormalFightPlan
from autowsgr.game.build import BuildManager
from autowsgr.game.game_operation import (
    ChangeShip,
    DestroyShip,
    MoveTeam,
    detect_ship_stats,
    quick_repair,
)
from autowsgr.ocr.ship_name import _recognize_ship
from autowsgr.port.common import Port, Ship
from autowsgr.port.ship import Fleet
from autowsgr.timer.timer import Timer
from autowsgr.utils.api_image import absolute_to_relative, crop_rectangle_relative
from autowsgr.utils.io import yaml_to_dict


class Task:
    def __init__(self, timer: Timer) -> None:
        self.port = timer.port
        self.timer = timer

    def run(self):
        """执行任务
        Returns:
            bool: 任务是否结束, 如果任务正常结束, 则从 TaskRunner 中删除
            list:
        """


class FightTask(Task):
    def __init__(self, timer: Timer, file_path="", *args, **kwargs) -> None:
        """
        Args:
            banned_ship (list(list(str))): 1-index 的列表, banned_ship[i] 表示第 i 号位不允许的舰船

            default_level_limit: 默认等级限制(2-111 之间的整数)

            level_limit (dict): 舰船名到等级限制的映射

            default_repair_mode: 默认修理方式

            repair_mode (dict): 舰船名到修理方式的映射

            ship_count: 舰队舰船数量

            fleet_id: 使用的舰队编号, 忽视 fight_plan 中的对应参数, 不支持第一舰队

            plan_path: 战斗策略文件位置, 参考 NormalFightPlan 的规范

            times: 任务重复次数

            max_repair_time: 最大修理时间: 超出该时间使用快修

            quick_repair: 无轮换时是否使用快修

            destroy_ship_types: 解装舰船

            all_ships: 所有参与轮换的舰船
        """
        super().__init__(timer)
        self.quick_repair = False
        self.destroy_ship_types = None
        self.default_level_limit = 2
        self.level_limit = {}
        self.default_repair_mode = 2
        self.repair_mode = {}
        self.ship_count = 0
        self.banned_ship = [[]] * 7
        self.fleet_id = 2
        self.times = 1
        self.plan_path = ""
        self.all_ships = []
        if file_path != "":
            self.__dict__.update(yaml_to_dict(file_path))
        self.__dict__.update(kwargs)
        if any([not self.port.have_ship(ship) for ship in self.all_ships]):
            self.timer.logger.info("含有未注册的舰船, 正在注册中...")

            # 添加到舰船名字列表中
            for ship in self.all_ships:
                if ship not in self.timer.ship_names:
                    self.timer.ship_names.append(ship)

            # 检查当前舰队中是否含有未初始化的舰船并将其初始化
            self.timer.goto_game_page("fight_prepare_page")
            fleet = Fleet(self.timer, self.fleet_id)
            fleet.detect()
            if any([ship in self.all_ships for ship in fleet.ships]):
                self.timer.logger.info("该舰队中处于修复状态的舰船正在被快速修复")
                quick_repair(self.timer, 3, switch_back=True)

            for i, ship in enumerate(fleet.ships):
                if ship in self.all_ships and not self.port.have_ship(ship):
                    tmp = self.port.register_ship(ship)
                    tmp.level = fleet.levels[i]
                    tmp.statu = detect_ship_stats(self.timer)[i]

            # 逐个初始化舰船, 效率较低, 待优化
            for ship in self.all_ships:
                if not self.port.have_ship(ship):  #
                    self.timer.logger.info(f"正在尝试注册 {ship}")
                    try:
                        ChangeShip(self.timer, self.fleet_id, 1, ship)
                        quick_repair(timer, 3, switch_back=True)
                    except ShipNotFoundErr as e:
                        self.timer.relative_click(0.05, 0.05)
                        self.timer.logger.warning(f"舰船 {ship} 注册失败, 放弃注册")
                        continue
                        # raise BaseException(f"未找到 {ship} 舰船")
                    tmp = self.port.register_ship(ship)
                    tmp.statu = detect_ship_stats(self.timer)[1]
                    fleet.detect()
                    tmp.level = fleet.levels[1]
            self.port.show_fleet()

    def build_fleet(self, ignore_statu=False):
        """尝试组建出征舰队"""

        def _build_fleet(ships):
            fleet = [
                None,
            ]
            ships = [ship for ship in ships if self.port.have_ship(ship)]
            ships = set(ships)
            for i in range(1, self.ship_count + 1):
                fleet.append(None)
                for ship in ships:
                    if self.banned_ship == -1 or ship not in self.banned_ship[i]:
                        fleet[i] = ship
                        ships.remove(ship)
                        break

            return (
                None
                if any([ship is None for ship in fleet[1 : self.ship_count + 1]])
                else fleet
            )

        if self.ship_count not in range(1, 7):
            raise ValueError(
                f"舰队舰船数量设置错误或者未设置, 当前设置的舰船数量为 {self.ship_count}"
            )
        # 清除已经达到等级要求的舰船并检查
        self.all_ships = [
            ship
            for ship in self.all_ships
            if self.port.get_ship_by_name(ship).level
            < self.level_limit.get(ship, self.default_level_limit)
        ]
        fleet = _build_fleet(self.all_ships)
        if fleet is None:
            self.timer.logger.info("由于等级超限无法组织舰队, 任务终止.")
            return 1, None

        # 清除不满足出征条件的舰船并检查
        fleet = _build_fleet(
            [
                ship
                for ship in self.all_ships
                if (
                    (
                        ignore_statu == False
                        and self.port.get_ship_by_name(ship).statu
                        < self.repair_mode.get(ship, self.default_repair_mode)
                    )
                    or (ignore_statu == True)
                )
            ]
        )
        if fleet is None:
            self.timer.logger.info(
                "由于部分舰船不满足出征条件而无法组织舰队, 任务暂停中..."
            )
            return 2, None
        return 0, fleet

    def check_repair(self):
        tasks = []
        for name in self.all_ships:
            ship = self.port.get_ship_by_name(name)
            if ship.statu != 3 and ship.statu >= self.repair_mode.get(
                ship, self.default_repair_mode
            ):
                # 满足修理条件
                if ship.waiting_repair:
                    self.timer.logger.debug(
                        f"舰船 {name} 已在修理队列中, 不再重复添加."
                    )
                else:
                    ship.waiting_repair = True
                    self.timer.logger.info(f"添加舰船 {name} 到修理队列.")
                    tasks.append(RepairTask(self.timer, ship))
        return tasks

    def run(self):
        # 应该退出的情况: 1.等级限制 2.不允许快修, 需要等待 3. 船坞已满, 需清理
        statu, fleet = self.build_fleet()
        if statu == 1:
            return True, []
        if statu == 2 and not self.quick_repair:
            return False, self.check_repair() + [self]
        if self.port.ship_factory.full:
            tasks = [
                OtherTask(
                    self.timer, "destroy", destroy_ship_types=self.destroy_ship_types
                )
            ]
            return False, tasks

        plan = NormalFightPlan(self.timer, self.plan_path, self.fleet_id, fleet=fleet)
        plan.repair_mode = [3] * 6
        # 设置战时快修
        if statu == 2:
            statu, fleet = self.build_fleet(True)
            for i, name in enumerate(fleet):
                if name == None:
                    continue
                ship = self.port.get_ship_by_name(name)
                if ship.statu >= self.repair_mode.get(name, self.default_repair_mode):
                    self.timer.logger.info(f"舰船 {name} 的状态已经标记为修复")
                    plan.repair_mode[i] = ship.statu
                    ship.set_repair(0)
        # 执行战斗
        ret = plan.run()
        # 处理船坞已满
        if ret == literals.DOCK_FULL_FLAG:
            return False, [
                OtherTask(
                    self.timer, "destroy", destroy_ship_types=self.destroy_ship_types
                )
            ]
        self.times -= 1
        # 更新舰船状态
        self.timer.wait_pages("map_page")
        self.timer.goto_game_page("fight_prepare_page")
        MoveTeam(self.timer, self.fleet_id)
        fleet = Fleet(self.timer)
        fleet.detect()
        ship_stats = detect_ship_stats(self.timer)
        for i, name in enumerate(fleet.ships):
            if name == None:
                continue
            ship = self.port.get_ship_by_name(name)
            if ship is not None:
                ship.level = fleet.levels[i]
                ship.statu = ship_stats[i]

        return True, self.check_repair() + [self]


class BuildTask(Task):
    def __init__(self, port) -> None:
        super().__init__(port)


class RepairTask(Task):
    def __init__(self, timer: Timer, ship: Ship, *args, **kwargs) -> None:
        super().__init__(timer)
        self.max_repiar_time = 1e9
        self.__dict__.update(kwargs)
        self.ship = ship

    def run(self):
        def switch_quick_repair(enable: bool):
            enabled = self.timer.check_pixel(
                (445, 91), (253, 150, 40), screen_shot=True
            )
            if enabled != enable:
                self.timer.relative_click(0.464, 0.168)

        waiting = self.port.bathroom.get_waiting_time()
        if waiting == None:
            # 检查等待时间
            available_time = []
            timer = self.timer
            timer.goto_game_page("bath_page")
            baths = [(0.076, 0.2138), (0.076, 0.325), (0.076, 0.433)]
            repair_position = [
                (0.283, 0.544),
                (0.530, 0.537),
                (0.260, 0.656),
                (0.513, 0.644),
            ]

            for i in range(self.timer.config.bath["bathroom_feature_count"]):
                timer.relative_click(*baths[i])
                for j in range(4):
                    timer.relative_click(*repair_position[j])
                    time.sleep(0.5)
                    timer.update_screen()
                    timer.logger.info(f"检查中:{(i, j)}")
                    if "快速修理" in [
                        result[1]
                        for result in timer.recognize_screen_relative(
                            0.279, 0.319, 0.372, 0.373
                        )
                    ]:
                        timer.logger.info(f"此位置有舰船修理中")
                        seconds = self.port.bathroom._time_to_seconds(
                            timer.recognize_screen_relative(0.413, 0.544, 0.506, 0.596)[
                                0
                            ][1]
                        )
                        timer.logger.info(f"预期用时: {seconds} 秒")
                        available_time.append(time.time() + seconds)
                        timer.relative_click(
                            797 / 1280, 509 / 720, delay=1
                        )  # 关闭快修选择界面

            while len(available_time) < timer.config.bath["bathroom_count"]:
                available_time.append(0)
            info = str(
                [
                    datetime.datetime.fromtimestamp(timestamp).strftime(
                        "%Y-%m-%d %H:%M:%S"
                    )
                    for timestamp in available_time
                ]
            )
            self.port.bathroom.update_available_time(available_time)
            timer.logger.info(f"浴室信息检查完毕:{info}")

        # 扫描等待修理列表
        last_result = None
        self.timer.goto_game_page("choose_repair_page")
        while True:
            time_costs = self.timer.recognize_screen_relative(
                0.041, 0.866, 0.966, 0.943, True
            )
            for time_cost in time_costs:
                self.timer.update_screen()
                text = time_cost[1].replace(" ", "")
                if text.startswith("耗时") and len(text) == 11:
                    # 整个图像截取完全
                    x = 0.041 + time_cost[0][0][0] / self.timer.screen.shape[1]
                    y = 0.741
                    img = crop_rectangle_relative(self.timer.screen, x, y, 0.12, 0.042)
                    name = _recognize_ship(img, self.timer.ship_names)
                    if len(name) == 0:
                        # 单字船名识别失败
                        continue
                    name = name[0][0]
                    seconds = self.port.bathroom._time_to_seconds(text[3:])
                    if name == self.ship.name:
                        if self.max_repiar_time <= seconds:
                            # 快速修复
                            self.timer.logger.info("满足快速修复条件, 使用快速修复工具")
                            self.ship.set_repair(0)
                            switch_quick_repair(True)
                            self.timer.relative_click(x, y)
                            return True, []

                        if not self.port.bathroom.is_available():
                            self.timer.logger.info(
                                "当前浴场已满, 不允许快速修复, 此任务延后"
                            )
                            return False, []

                        self.port.bathroom.add_repair(text[3:])
                        self.timer.relative_click(x, y)
                        self.timer.set_page("bath_page")
                        self.ship.set_repair(seconds)
                        self.port.bathroom.add_repair(text[3:])
                        return True, []

                    else:
                        self.timer.logger.debug(f"识别到舰船: {name}")

            self.timer.relative_swipe(0.33, 0.5, 0.66, 0.5, delay=1)
            time.sleep(0.5)
            if time_costs == last_result:
                raise BaseException("未找到目标舰船")
            last_result = time_costs


class OtherTask(Task):
    def __init__(self, timer: Timer, type, *args, **kwargs) -> None:
        """其它类型的任务
        Args:
            type (str): 任务类型
                "destroy": 舰船解装
                "empty": 不执行任何操作
                "build": 建造舰船
                "develop": 开发装备
            use_quick_build (bool): 是否允许使用快速修理
            resources_use (tuple): 按顺序油弹钢铝四项
        """
        super().__init__(timer)
        self.type = type
        if type == "destroy":
            self.destroy_ship_types = kwargs["destroy_ship_types"]
            timer.logger.info("船舱已满, 添加解装任务中...")
            if timer.port.ship_factory.waiting_destory:
                timer.logger.info("任务队列中已经有解装任务, 跳过")
                self.run = lambda self: None

        if type == "build" or type == "develop":
            try:
                self.resources_use = kwargs["resources_use"]
                if "use_quick_build" in kwargs:
                    self.use_quick_build = kwargs["use_quick_build"]
                else:
                    timer.logger.warning("未指定是否使用快修, 默认不使用")
                    self.use_quick_build = False
            except:
                raise ValueError("未指定油弹钢铝数量")

    def run(self):
        if self.type == "destroy":
            DestroyShip(self.timer, ship_types=self.destroy_ship_types)

        if self.type == "build" or self.type == "develop":
            factory: BuildManager = self.timer.port.factory
            type = "ship" if self.type == "build" else "equipment"
            factory.update_slot_eta(type)
            if factory.has_empty_slot(type):
                factory.build(type, self.resources_use)
            elif not self.use_quick_build:
                factory.build(type, self.resources_use, True)
            else:
                return False, []
        return True, []


class TaskRunner:
    def __init__(self) -> None:
        self.tasks = []
        pass

    def run(self):
        while True:
            # 调度逻辑: 依次尝试任务列表中的每个任务, 如果任务正常结束, 则从头开始, 否则找下一个任务.
            # 每个任务尝试执行后, 会向任务列表中添加一些新任务.
            id = 0
            while id < len(self.tasks):
                task = self.tasks[id]
                statu, new_tasks = task.run()
                if statu:
                    self.tasks = self.tasks[0:id] + new_tasks + self.tasks[id + 1 :]
                    break
                else:
                    self.tasks = (
                        self.tasks[0 : id + 1] + new_tasks + self.tasks[id + 1 :]
                    )
                    id += 1



import time

from autowsgr.timer.timer import Timer, try_to_get_expedition


class Expedition:
    def __init__(self, timer: Timer) -> None:
        self.timer = timer
        self.is_ready = False
        self.last_check = time.time()

    def update(self, force=False):
        self.timer.update_screen()
        if (
            isinstance(self.timer.now_page, str) and "unknow" in self.timer.now_page
        ) or self.timer.now_page.name not in [
            "expedition_page",
            "map_page",
            "battle_page",
            "exercise_page",
            "decisive_battle_entrance",
        ]:
            if force or time.time() - self.last_check > 1800:
                self.timer.goto_game_page("main_page")
            if self.timer.now_page.name == "main_page":
                self.is_ready = self.timer.check_pixel(
                    (933, 454), bgr_color=(45, 89, 255)
                )
        else:
            self.is_ready = self.timer.check_pixel((464, 11), bgr_color=(45, 89, 255))

    def run(self, force=False):
        """检查远征, 如果有未收获远征, 则全部收获并用原队伍继续

        Args:
            force (bool): 是否强制检查
        Returns:
            bool: 是否进行了远征操作
        """
        self.update(force=force)
        if self.is_ready:
            self.timer.goto_game_page("expedition_page")
            flag = try_to_get_expedition(self.timer)
            self.timer.last_expedition_check_time = time.time()



import datetime
import os
import subprocess
import winreg
from types import SimpleNamespace

import keyboard as kd
from airtest.core.error import AdbError

import autowsgr
from autowsgr.constants.data_roots import TUNNEL_ROOT
from autowsgr.game.build import BuildManager
from autowsgr.timer import Timer
from autowsgr.utils.io import recursive_dict_update, yaml_to_dict
from autowsgr.utils.logger import Logger
from autowsgr.utils.update import check_for_updates

event_pressed = set()
script_end = 0


def initialize_logger_and_config(settings_path):
    config = yaml_to_dict(
        os.path.join(
            os.path.dirname(autowsgr.__file__), "data", "default_settings.yaml"
        )
    )
    if settings_path is not None:
        user_settings = yaml_to_dict(settings_path)
        config = recursive_dict_update(config, user_settings)
    else:
        print("========Warning========")
        print(
            f"No user_settings file specified, default settings "
            f"{os.path.join(os.path.dirname(autowsgr.__file__), 'data', 'default_settings.yaml')}"
            f" will be used."
        )
        print("=========End===========")

    # reading the registry for emulator if needed
    if config["emulator"]["start_cmd"] == "":
        print("========Warning========")
        print("No emulator directory provided, reading the registry")
        config["emulator"]["start_cmd"] = get_emulator_path(config["emulator"]["type"])
        print("The emulator directory is " + config["emulator"]["start_cmd"])
        print("=========End===========")

    # set logger
    config["log_dir"] = os.path.join(
        config["LOG_PATH"], datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    )
    os.makedirs(config["log_dir"], exist_ok=True)
    logger = Logger(config)
    config = SimpleNamespace(**config)
    if config.check_update:
        try:
            check_for_updates()
        except Exception as e:
            print(f"Failed to check for updates: {e}")
    config_str = logger.save_config(config)
    logger.reset_level()
    return config, logger


def start_script(settings_path=None):
    """启动脚本, 返回一个 Timer 记录器.
    :如果模拟器没有运行, 会尝试启动模拟器,
    :如果游戏没有运行, 会自动启动游戏,
    :如果游戏在后台, 会将游戏转到前台
    Returns:
        Timer: 该模拟器的记录器
    """
    # set logger
    try:
        config, logger = initialize_logger_and_config(settings_path)
        timer = Timer(config, logger)
        timer.port.factory = BuildManager(timer)
    except AdbError:
        adb_exe = os.path.join(os.path.dirname(TUNNEL_ROOT), "adb", "adb.exe")
        subprocess.run([adb_exe, "devices", "-l"])
        logger.warning("Adb 连接模拟器失败, 正在清除原有连接并重试")
        timer = Timer(config, logger)
    return timer


def get_emulator_path(emulator_type):
    try:
        if emulator_type == "雷电":
            key = winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\leidian\ldplayer9"
            )
            try:
                path, _ = winreg.QueryValueEx(key, "InstallDir")
                return os.path.join(path, "dnplayer.exe")
            except FileNotFoundError:
                print("Path not found")
            finally:
                winreg.CloseKey(key)

        elif emulator_type == "蓝叠 Hyper-V":
            key = winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE, r"SOFTWARE\BlueStacks_nxt_cn"
            )
            try:
                path, _ = winreg.QueryValueEx(key, "InstallDir")
                return os.path.join(path, "HD-Player.exe")
            except FileNotFoundError:
                print("Path not found")
            finally:
                winreg.CloseKey(key)

    except FileNotFoundError:
        print("Emulator not found")
        return None



import json
import os
import sys
from datetime import datetime

from loguru import logger

from autowsgr.utils.io import save_image


class Logger:
    def __init__(self, config):
        self.config = config
        self.log_dir = config["log_dir"]
        if "log_level" in config.keys():
            log_level = config["log_level"]
        else:
            log_level = "INFO"
        self.log_level = log_level

    def _configure_logger(self, log_level):

        # 自定义格式化器
        def custom_format(record):
            level = record["level"].name
            message = record["message"]

            # 绿色：32，红色：31，黄色：33，蓝色：34，淡蓝色：36，清除：0
            if level == "INFO":
                return "\x1b[32m{time:HH:mm:ss} \x1b[36mautowsgr \x1b[32m{level:^7}\x1b[0m| {message}\n"
            elif level == "ERROR":
                return "\x1b[32m{time:HH:mm:ss} \x1b[36mautowsgr \x1b[31m{level:^7}\x1b[0m| {message}\n"
            elif level == "DEBUG":
                return "\x1b[32m{time:HH:mm:ss} \x1b[36mautowsgr \x1b[0m{level:^7}| {message}\n"
            elif level == "WARNING":
                return "\x1b[32m{time:HH:mm:ss} \x1b[36mautowsgr \x1b[33m{level:^7}\x1b[0m| {message}\n"
            else:
                return "\x1b[32m{time:HH:mm:ss} \x1b[36mautowsgr \x1b[31m{level:^7}| {message}\n"

        # 添加文件日志记录器
        logger.add(
            os.path.join(self.log_dir, "log_{time}.log"),
            level=log_level,
            colorize=True,
            format="{time:HH:mm:ss} autowsgr {level:^7}| {message}",
            catch=True,
            retention="1 week",
        )

        # 添加控制台日志记录器
        # logger.add(
        #    sys.stdout,
        #    level=log_level,
        #    colorize=True,
        #    format="<green>{time:HH:mm:ss}</green> <level>{level}</level> <cyan>autowsgr:</cyan> <level>{message}</level>",
        #    catch=True,
        # )

        # 添加控制台日志记录器
        logger.add(
            lambda msg: print(msg, end=""),
            level=log_level,
            colorize=True,
            format=custom_format,
            catch=True,
        )

    def save_config(self, config):
        # write config file
        config_str = json.dumps(
            vars(config), ensure_ascii=False, indent=4, sort_keys=True
        )
        with open(
            os.path.join(self.log_dir, "config.json"), "w", encoding="utf-8"
        ) as f:
            f.write(config_str)
        return config_str

    def reset_level(self):
        logger.remove()
        self._configure_logger(self.log_level)

    def debug(self, *args):
        logger.debug(str(args))

    def info(self, string):
        logger.info(string)
        # st.write(string)

    def warning(self, string):
        logger.warning("===================WARNING===================")
        logger.warning(string)
        logger.warning("====================END====================")
        # st.write(string)

    def error(self, string):
        logger.error("===================ERROR===================")
        logger.error(string)
        logger.error("====================END====================")

    def log_image(
        self,
        image,
        name=None,
        ndarray_mode="BGR",
        ignore_existed_image=False,
        *args,
        **kwargs,
    ):
        """向默认数据记录路径记录图片
        Args:
            image: 图片,PIL.Image.Image 格式或者 numpy.ndarray 格式
            name (str): 图片文件名
        """
        if name is None:
            name = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        if "png" not in name and "PNG" not in name:
            name += ".PNG"
        path = os.path.join(self.log_dir, name)

        save_image(
            path=path,
            image=image,
            ignore_existed_image=ignore_existed_image,
            *args,
            **kwargs,
        )
        self.info(f"图片已保存至{path}")



import os
import subprocess
from typing import List, Tuple

import cv2
import numpy as np
from thefuzz import process

from autowsgr.constants.data_roots import TUNNEL_ROOT


def edit_distance(word1, word2) -> int:
    """
    解题思路，动态规划
        步骤1：将word1与word2前拼接上空格，方便为空时的操作
        步骤2：初始化dp第一个元素，接着初始化dp的第一行与第一列
        步骤3：可通过画表(如: excel里)找到状态转移的规律，填充剩下的dp格子即可
    :return: dp[-1][-1], 返回最后操作的结果
    """
    m, n = len(word1), len(word2)
    if m == 0 and n == 0:
        return 0
    word1, word2 = (
        " " + word1,
        " " + word2,
    )  # 非常必要的操作，不添加空格话，在Word为空时会比较麻烦
    dp = [[0 for _ in range(n + 1)] for _ in range(m + 1)]
    dp[0][0] = 0  # 初始化dp[0][0] = 0，因为空格对空格不需要任何操作，即0步
    for i in range(1, n + 1):  # 第一行初始化
        dp[0][i] = i
    for i in range(1, m + 1):  # 第一列初始化
        dp[i][0] = i
    for i in range(1, m + 1):  # 逐个填充剩余的dp格子
        for j in range(1, n + 1):
            if word1[i] == word2[j]:
                dp[i][j] = dp[i - 1][j - 1]
            else:
                dp[i][j] = min(dp[i - 1][j - 1], dp[i - 1][j], dp[i][j - 1]) + 1
    return dp[-1][-1]


def find_lcseque(s1, s2):
    """求两个字符串的LCS"""
    m = [[0 for x in range(len(s2) + 1)] for y in range(len(s1) + 1)]
    d = [[None for x in range(len(s2) + 1)] for y in range(len(s1) + 1)]
    for p1 in range(len(s1)):
        for p2 in range(len(s2)):
            if s1[p1] == s2[p2]:
                m[p1 + 1][p2 + 1] = m[p1][p2] + 1
                d[p1 + 1][p2 + 1] = "ok"
            elif m[p1 + 1][p2] > m[p1][p2 + 1]:
                m[p1 + 1][p2 + 1] = m[p1 + 1][p2]
                d[p1 + 1][p2 + 1] = "left"
            else:
                m[p1 + 1][p2 + 1] = m[p1][p2 + 1]
                d[p1 + 1][p2 + 1] = "up"
    (p1, p2) = (len(s1), len(s2))
    s = []
    while m[p1][p2]:  # 不为None时
        c = d[p1][p2]
        if c == "ok":  # 匹配成功，插入该字符，并向左上角找下一个
            s.append(s1[p1 - 1])
            p1 -= 1
            p2 -= 1
        if c == "left":  # 根据标记，向左找下一个
            p2 -= 1
        if c == "up":  # 根据标记，向上找下一个
            p1 -= 1
    s.reverse()
    return "".join(s)


class OCRBackend:
    WORD_REPLACE = None  # 记录中文ocr识别的错误用于替换。主要针对词表缺失的情况，会导致稳定的识别为另一个字

    def __init__(self, config, logger):
        self.config = config
        self.logger = logger

    def read_text(
        self, img, allowlist: List[str] = None, sort: str = "left-to-right", **kwargs
    ):
        """识别文字的具体实现，返回字符串格式识别结果"""
        raise NotImplementedError

    def resize_image_proportionally(self, image, scale_factor):
        """
        等比例扩大图片。

        参数:
        image -- NumPy数组格式的图片
        scale_factor -- 扩大的倍数，例如2表示扩大到原来的两倍

        返回:
        resized_image -- 扩大后的图片
        """
        # 获取原始图片的尺寸
        original_height, original_width = image.shape[:2]

        # 计算新的尺寸
        new_width = int(original_width * scale_factor)
        new_height = int(original_height * scale_factor)

        # 使用cv2.resize函数等比例扩大图片
        resized_image = cv2.resize(
            image, (new_width, new_height), interpolation=cv2.INTER_LINEAR
        )
        return resized_image

    def recognize(
        self,
        img,
        allowlist: List[str] = None,
        candidates: List[str] = None,
        multiple=False,
        allow_nan=False,
        rgb_select=None,
        tolerance=30,
        scale_factor=2,
        **kwargs,
    ):
        """识别任意字符串, 该函数为最底层封装

        Args:
            scale_factor (int, optional): 图标缩放倍率. Defaults to 2.
        Returns:
            list((int, int), str, float): 中心位置, 字串识别结果, 置信度
        """

        def pre_process_rgb(img, rgb_select=None, tolerance=30):
            # 如果没有提供rgb_select，直接返回原始图像
            if rgb_select is None:
                return img

            # 将BGR图像转换为RGB格式
            img_rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

            rgb_select_normalized = np.array(rgb_select)
            mask = np.all(np.abs(img_rgb - rgb_select_normalized) <= tolerance, axis=-1)

            # 使用掩码将匹配的像素保留，其他像素设置为255
            result_img = img_rgb.copy()
            result_img[mask] = 0
            result_img[~mask] = 255

            # 将处理后的图像转换回BGR格式
            result_img_bgr = cv2.cvtColor(result_img, cv2.COLOR_RGB2BGR)

            return result_img_bgr

        def split_image(img: np.ndarray):
            def is_key_line(line: np.ndarray):
                return line.min() != 255
                # for i in range(1, line.shape[0], 8):
                #     if any(val != 255 for val in line[i]):
                #         return True
                # return False

            bias = []
            imgs = []
            last_array = np.ndarray([0, img.shape[1], 3], dtype=np.uint8)
            for i in range(img.shape[0]):
                if (not is_key_line(img[i])) or i == img.shape[0] - 1:
                    if last_array.shape[0] > 8:
                        imgs.append(last_array)
                        bias.append(i - last_array.shape[0])
                    last_array = np.ndarray([0, img.shape[1], 3], dtype=np.uint8)
                    # else:
                else:
                    last_array = np.append(last_array, [img[i]], 0)
            return imgs, bias

        def post_process_text(t):
            for k, v in self.WORD_REPLACE.items():
                t = t.replace(k, v)
            res, lcs, name = "dsjfagiahsdifhaoisd", "", t
            if candidates is None:
                return name
            for _name in candidates:
                if any([(_name.find(char) != -1) for char in name]):
                    dis1 = edit_distance(_name, name) / (
                        len(find_lcseque(_name, name)) + 1
                    )
                    dis2 = edit_distance(res, name) / (len(find_lcseque(res, name)) + 1)
                    if dis1 < dis2:
                        res = _name
            return res

        img = pre_process_rgb(img, rgb_select, tolerance)
        if type(img) == type("1234"):
            img = cv2.imread(img)
        img = self.resize_image_proportionally(img, scale_factor)
        imgs, bias = split_image(img)
        results = []
        for img, bia in zip(imgs, bias):
            _results = self.read_text(img, allowlist, **kwargs)
            _results = [x for x in _results if x[1] != ""]  # 去除空匹配
            _results = [(t[0], post_process_text(t[1]), t[2]) for t in _results]
            for i, result in enumerate(_results):
                results.append(
                    (
                        [
                            (result[0][0]) // scale_factor,
                            (result[0][1] + bia) // scale_factor,
                        ],
                        result[1],
                        result[2],
                    )
                )
        if self.config.SHOW_OCR_INFO:
            self.logger.debug(f"修正OCR结果：{results}")

        if allow_nan and not results:
            return None

        if multiple:
            return results
        else:
            if not results:
                results = ["Unkown"]
            return results[0]

    def recognize_number(
        self, img, extra_chars="", multiple=False, allow_nan=False, **kwargs
    ):
        """识别数字"""

        def process_number(t: str):
            # 今日胖次、掉落; 决战升级经验等
            if "/" in t:
                nums = t.split("/")
                assert len(nums) == 2
                return process_number(nums[0]), process_number(nums[1])

            # 决战，费用是f"x{cost}"格式
            t = t.lstrip("xXKk")
            # 战后经验值 f"Lv.{exp}"格式
            t = t.lstrip("Lv.")
            # 建造资源有前导0
            if t != "0":
                t = t.lstrip("0")

            # 资源可以是K/M结尾
            if t.endswith("K") or t.endswith("k"):
                return eval(t[:-1]) * 1000
            if t.endswith("M"):
                return eval(t[:-1]) * 1000000

            return eval(t)

        results = self.recognize(
            img, allowlist="0123456789" + extra_chars, multiple=True, **kwargs
        )
        results = [(t[0], process_number(t[1]), t[2]) for t in results]
        if self.config.SHOW_OCR_INFO:
            self.logger.debug(f"数字解析结果：{results}")

        if allow_nan and not results:
            return None

        if multiple:
            return results
        else:
            if not len(results) == 1:
                self.logger.warning(f"OCR识别数字失败: {results}")
                results = []
            return results[0]

    def recognize_ship(self, image, candidates, **kwargs):
        """传入一张图片,返回舰船信息,包括名字和舰船型号"""
        if isinstance(image, str):
            image_path = os.path.abspath(image)
        else:
            image_path = os.path.join(TUNNEL_ROOT, "OCR.PNG")
            cv2.imwrite(image_path, image)
        with open(os.path.join(TUNNEL_ROOT, "locator.in"), "w+") as f:
            f.write(image_path)
        locator_exe = os.path.join(TUNNEL_ROOT, "locator.exe")
        subprocess.run([locator_exe, TUNNEL_ROOT])
        if os.path.exists(os.path.join(TUNNEL_ROOT, "1.PNG")):
            img_path = os.path.join(TUNNEL_ROOT, "1.PNG")
        else:
            img_path = "1.PNG"
        return self.recognize(img_path, candidates=candidates, multiple=True, **kwargs)

    # def recognize_time(self, img, format="%H:%M:%S"):
    #     """识别时间"""
    #     text = self.recognize(img, allowlist="0123456789:").replace(" ", "")
    #     return str2time(text, format)


class EasyocrBackend(OCRBackend):
    WORD_REPLACE = {
        "鲍鱼": "鲃鱼",
        "鲴鱼": "鲃鱼",
    }

    def __init__(self, config, logger) -> None:
        super().__init__(config, logger)
        import easyocr

        self.reader = easyocr.Reader(["ch_sim", "en"])

    def read_text(
        self,
        img,
        allowlist: List[str] = None,
        sort="left-to-right",
        # TODO：以下参数可能需要调整，以获得最好OCR性能
        min_size=7,
        text_threshold=0.25,
        low_text=0.3,
        **kwargs,
    ):
        """识别文字的具体实现，返回字符串格式识别结果"""

        def get_center(pos1, pos2):
            x1, y1 = pos1
            x2, y2 = pos2
            return (x1 + x2) / 2, (y1 + y2) / 2

        results = self.reader.readtext(
            img,
            allowlist=allowlist,
            min_size=min_size,
            text_threshold=text_threshold,
            low_text=low_text,
            **kwargs,
        )
        results = [(get_center(r[0][0], r[0][2]), r[1], r[2]) for r in results]

        if sort == "left-to-right":
            results = sorted(results, key=lambda x: x[0][0])
        elif sort == "top-to-bottom":
            results = sorted(results, key=lambda x: x[0][1])
        else:
            raise ValueError(f"Invalid sort method: {sort}")

        if self.config.SHOW_OCR_INFO:
            self.logger.debug(f"原始OCR结果: {results}")
        return results


class PaddleOCRBackend(OCRBackend):
    WORD_REPLACE = {
        "鲍鱼": "鲃鱼",
    }

    def __init__(self, config, logger) -> None:
        super().__init__(config, logger)
        # TODO:后期单独训练模型，提高识别准确率，暂时使用现成的模型
        from paddleocr import PaddleOCR

        self.reader = PaddleOCR(
            use_angle_cls=True,
            use_gpu=True,
            show_log=False,
            lang="ch",
        )  # need to run only once to download and load model into memory

    def read_text(self, img, allowlist, **kwargs):

        def get_center(pos1, pos2):
            x1, y1 = pos1
            x2, y2 = pos2
            return (x1 + x2) / 2, (y1 + y2) / 2

        results = self.reader.ocr(img, cls=False, **kwargs)
        if results == [None]:
            results = []
        else:
            results = results[0]
        results = [(get_center(r[0][1], r[0][3]), r[1][0], r[1][1]) for r in results]
        if self.config.SHOW_OCR_INFO:
            self.logger.debug(f"原始OCR结果: {results}")
        return results



import os
import re
import subprocess
import time

import inquirer
import pkg_resources
import requests
from packaging.version import parse

update_source = [
    inquirer.List(
        "source",
        message="Please choose the source to update",
        choices=["清华源(推荐)", "北京外国语", "PyPI"],
    ),
]


def check_for_updates():
    print("Checking for updates...")
    # 获取本地autowsgr版本号
    local_version = get_local_version()

    # 发送 GET 请求获取库的元数据信息
    response = requests.get("https://pypi.tuna.tsinghua.edu.cn/pypi/autowsgr/json")
    data = response.json()

    # 提取最新版本号
    latest_version = data["info"]["version"]
    # 比较版本号
    if parse(local_version) < parse(latest_version):
        update_questions = [
            inquirer.List(
                "source",
                message=f"New version {latest_version} is available.Your version is {local_version}. Do you want to update?",
                choices=["Yes", "No"],
            ),
        ]
        result = get_user_choice(update_questions)
        if result == "Yes":
            # 选择使用哪个源更新,输出按钮回车选择
            choice = get_user_choice(update_source)
            update_library(choice)
            recent_updates = get_recent_updates_from_pypi(latest_version)
            print("更新内容:\n" + recent_updates)

            print("更新完成，稍后将自动退出，请重新启动脚本")
            time.sleep(5)
            os._exit(0)  # 更新成功后退出脚本
    else:
        print("You are using the latest version of the library.")


def get_local_version():
    # 使用pkg_resources获取本地库的版本号
    try:
        version = pkg_resources.get_distribution("autowsgr").version
        return version
    except Exception as e:
        print(f"Failed to get the local version.Error: {e}")
        return


def get_user_choice(questions):
    answers = inquirer.prompt(questions)
    return answers["source"]


def update_library(choice="PyPI"):
    choice_list = {
        "PyPI": [
            "pip",
            "install",
            "--upgrade",
            "--index-url",
            "https://pypi.org/simple",
            "autowsgr",
        ],
        "北京外国语": [
            "pip",
            "install",
            "--index-url",
            "https://mirrors.bfsu.edu.cn/pypi/web/simple/",
            "--upgrade",
            "autowsgr",
        ],
        "清华源(推荐)": [
            "pip",
            "install",
            "--index-url",
            "https://pypi.tuna.tsinghua.edu.cn/simple",
            "--upgrade",
            "autowsgr",
        ],
    }
    subprocess.run(choice_list[choice])


def get_recent_updates_from_pypi(latest_version):

    url = f"https://pypi.org/project/autowsgr/{latest_version}/#description"
    response = requests.get(url)

    if response.status_code == 200:
        readme_content = response.text
        updates_section = re.search(
            r"<h2>近期更新</h2>(.*?)</ul>", readme_content, re.S
        )

        if updates_section:
            updates = updates_section.group(1).strip()
            # 提取所有 <li> 标签中的内容
            updates_list = re.findall(r"<li>(.*?)</li>", updates, re.S)
            # 合并为一个字符串，每行前面加上一个 ·
            updates_text = "\n".join(
                [f'· {re.sub(r"<.*?>", "", update).strip()}' for update in updates_list]
            )
            return updates_text
        else:
            return "未找到近期更新部分。"
    else:
        return f"无法获取更新内容，状态码: {response.status_code}"


# if __name__ == "__main__":
#     check_for_updates()



from functools import partial

from airtest.core.cv import (
    MATCHING_METHODS,
    ST,
    InvalidMatchingMethodError,
    TargetPos,
    Template,
)
from airtest.core.settings import Settings as ST

from autowsgr.constants.data_roots import IMG_ROOT
from autowsgr.utils.io import create_namespace


class MyTemplate(Template):
    def __radd__(self, other):
        if isinstance(other, list):
            return other + [self]  # 添加到列表开头
        else:
            return NotImplemented

    def __add__(self, other):
        if isinstance(other, list):
            return [self] + other  # 添加到列表末尾
        else:
            return NotImplemented

    def match_in(self, screen, this_methods=None):
        match_result = self._cv_match(screen, this_methods)
        if not match_result:
            return None
        focus_pos = TargetPos().getXY(match_result, self.target_pos)
        return focus_pos

    def _cv_match(self, screen, this_methods=None):
        ori_image = self._imread()
        image = self._resize_image(ori_image, screen, ST.RESIZE_METHOD)
        ret = None
        if this_methods is None:
            this_methods = ST.CVSTRATEGY
        for method in this_methods:
            # get function definition and execute:
            func = MATCHING_METHODS.get(method, None)
            if func is None:
                raise InvalidMatchingMethodError(
                    f"Undefined method in CVSTRATEGY: '{method}', try 'kaze'/'brisk'/'akaze'/'orb'/'surf'/'sift'/'brief' instead."
                )
            if method in ["mstpl", "gmstpl"]:
                ret = self._try_match(
                    func,
                    ori_image,
                    screen,
                    threshold=self.threshold,
                    rgb=self.rgb,
                    record_pos=self.record_pos,
                    resolution=self.resolution,
                    scale_max=self.scale_max,
                    scale_step=self.scale_step,
                )
            else:
                ret = self._try_match(
                    func, image, screen, threshold=self.threshold, rgb=self.rgb
                )
            if ret:
                break
        return ret


IMG = create_namespace(
    IMG_ROOT, partial(MyTemplate, threshold=0.9, resolution=(960, 540))
)



import time

from autowsgr.constants.custom_exceptions import ImageNotFoundErr, ShipNotFoundErr
from autowsgr.constants.image_templates import IMG
from autowsgr.constants.other_constants import SHIP_TYPE_CLICK
from autowsgr.constants.positions import BLOOD_BAR_POSITION
from autowsgr.game.get_game_info import check_support_stats, detect_ship_stats
from autowsgr.ocr.ship_name import recognize_ship
from autowsgr.timer import Timer
from autowsgr.utils.api_image import (
    absolute_to_relative,
    crop_image,
    crop_rectangle_relative,
)


def get_ship(timer: Timer):
    """获取掉落舰船"""

    def recognize_get_ship(timer: Timer):
        """识别获取 舰船/装备 页面斜着的文字，对原始图片进行旋转裁切"""
        NAME_POSITION = [(0.754, 0.268), (0.983, 0.009), 25]
        ship_name = timer.recognize(
            crop_image(timer.screen, *NAME_POSITION), candidates=timer.ship_names
        )[1]

        TYPE_POSITION = [(0.804, 0.27), (0.881, 0.167), 25]
        ship_type = timer.recognize(crop_image(timer.screen, *TYPE_POSITION))[1]

        return ship_name, ship_type

    ship_name = None
    ship_type = None
    timer.got_ship_num += 1
    if timer.port.ship_factory.capacity is not None:
        timer.logger.info(
            f"当前船坞容量 {timer.port.ship_factory.occupation}/{timer.port.ship_factory.capacity}"
        )
        timer.port.ship_factory.occupation += 1
    while timer.wait_image([IMG.symbol_image[8]] + [IMG.symbol_image[13]], timeout=1):
        try:
            ship_name, ship_type = recognize_get_ship(timer)
        except Exception as e:
            print(e)
        timer.click(915, 515, delay=0.25, times=1)
        timer.ConfirmOperation()
    timer.logger.info(f"获取舰船: {ship_name} {ship_type}")
    return ship_name, ship_type


def match_night(timer: Timer, is_night):
    """匹配夜战按钮并点击"""
    timer.wait_image(IMG.fight_image[6])
    while timer.wait_image(IMG.fight_image[6], timeout=0.5):
        if is_night:
            timer.click(325, 350, delay=0.5, times=1)
        else:
            timer.click(615, 350, delay=0.5, times=1)


def click_result(timer: Timer, max_times=1):
    """点击加速两页战果界面"""
    timer.wait_images(IMG.fight_image[14])
    while timer.wait_image(IMG.fight_image[14], timeout=0.5):
        timer.click(915, 515, delay=0.25, times=1)


def DestroyShip(timer: Timer, ship_types=None):
    """解装舰船，目前仅支持：全部解装+保留装备"""

    timer.go_main_page()
    timer.goto_game_page("destroy_page")
    timer.set_page("destroy_page")

    timer.click(90, 206, delay=1.5)  # 点添加

    # 选择舰船类型
    if ship_types is not None:
        timer.relative_click(0.912, 0.681)
        for ship_type in ship_types:
            timer.relative_click(
                *absolute_to_relative(SHIP_TYPE_CLICK[ship_type], (1280, 720)),
                delay=0.8,
            )
        timer.relative_click(0.9, 0.85, delay=1.5)

    timer.relative_click(0.91, 0.3, delay=1.5)  # 快速选择
    timer.relative_click(0.915, 0.906, delay=1.5)  # 确定
    timer.relative_click(0.837, 0.646, delay=1.5)  # 卸下装备
    timer.relative_click(0.9, 0.9, delay=1.5)  # 解装
    timer.relative_click(0.38, 0.567, delay=1.5)  # 四星确认

    # 识别船坞容量
    timer.click(90, 206, delay=1.5)  # 点添加
    # occupation, capacity = (crop_rectangle_relative(timer.get_screen(), 0.873, 0.035, 0.102, 0.038))
    # timer.port.ship_factory.update_capacity(capacity, occupation)
    # timer.logger.info(f"舰船容量: {occupation}/{capacity}")
    timer.go_main_page()


def verify_team(timer: Timer):
    """检验目前是哪一个队伍(1~4)
    含网络状况处理
    Args:
        timer (Timer): _description_

    Raises:
        ImageNotFoundErr: 未找到队伍标志
        ImageNotFoundErr: 不在相关界面

    Returns:
        int: 队伍编号(1~4)
    """
    if timer.identify_page("fight_prepare_page") == False:
        raise ImageNotFoundErr("not on fight_prepare_page")

    for _ in range(5):
        for i, position in enumerate([(64, 83), (186, 83), (310, 83), (430, 83)]):
            if timer.check_pixel(position, bgr_color=(228, 132, 16)):
                return i + 1
        time.sleep(0.2)
        timer.update_screen()

    if timer.process_bad_network():
        return verify_team(timer)

    timer.log_screen()
    raise ImageNotFoundErr()


def MoveTeam(timer: Timer, target, try_times=0):
    """切换队伍
    Args:
        timer (Timer): _description_
        target (_type_): 目标队伍
        try_times: 尝试次数
    Raise:
        ValueError: 切换失败
        ImageNotFoundErr: 不在相关界面
    """
    if try_times > 3:
        raise ValueError("can't change team")

    if timer.identify_page("fight_prepare_page") == False:
        timer.log_screen()
        raise ImageNotFoundErr("not on 'fight_prepare_page' ")

    if verify_team(timer) == target:
        return
    timer.logger.info("正在切换队伍到:" + str(target))
    timer.click(110 * target, 81)
    if verify_team(timer) != target:
        MoveTeam(timer, target, try_times + 1)


def SetSupport(timer: Timer, target, try_times=0):
    """启用战役支援
    Args:
        timer (Timer): _description_
        target (bool, int): 目标状态
    Raise:
        ValueError: 未能成功切换战役支援状态
    """
    target = bool(target)
    timer.goto_game_page("fight_prepare_page")

    if check_support_stats(timer) != target:
        timer.click(628, 82, delay=1)
        timer.click(760, 273, delay=1)
        timer.click(480, 270, delay=1)
        timer.logger.info("开启支援状态成功")

    if timer.is_bad_network(0) or check_support_stats(timer) != target:
        if timer.process_bad_network("set_support"):
            SetSupport(timer, target)
        else:
            raise ValueError("can't set right support")


def quick_repair(
    timer: Timer, repair_mode=2, ship_stats=None, switch_back=False, *args, **kwargs
):
    """战斗界面的快速修理
    Args:
        timer (Timer): _description_
        repair_mode:
            1: 快修, 修中破
            2: 快修, 修大破
            3: 快修, 修复所有正在修理中的船
    """
    arg = repair_mode
    try:
        if ship_stats == None:
            ship_stats = detect_ship_stats(timer)
        if not any(x in ship_stats for x in [0, 1, 2]):
            time.sleep(1)
            ship_stats = detect_ship_stats(timer)
        if not any(x in ship_stats for x in [0, 1, 2]):
            timer.logger.warning("执行修理操作时没有成功检测到舰船")
            raise ValueError("没有成功检测到舰船，请检查是否正确编队")

        assert type(repair_mode) in [int, list, tuple]
        if type(repair_mode) == int:  # 指定所有统一修理方案
            repair_mode = [repair_mode for _ in range(6)]

        assert len(repair_mode) == 6
        need_repair = [False for _ in range(6)]
        for i, x in enumerate(repair_mode):
            assert x in [1, 2, 3]
            if x == 1:
                need_repair[i] = ship_stats[i + 1] not in [-1, 0]
            elif x == 2:
                need_repair[i] = ship_stats[i + 1] not in [-1, 0, 1]
            else:
                need_repair[i] = False

        if timer.config.DEBUG:
            timer.logger.debug("ship_stats:", ship_stats)
        if any(need_repair) or timer.image_exist(IMG.repair_image[1]):
            timer.click(420, 420, times=2, delay=0.8)  # 进入修理页面
            # 快修已经开始泡澡的船
            pos = timer.get_image_position(IMG.repair_image[1])
            while pos != None:
                timer.port.bathroom.available_time = None
                timer.click(pos[0], pos[1], delay=1)
                pos = timer.get_image_position(IMG.repair_image[1])
            # 按逻辑修理
            for i in range(1, 7):
                if need_repair[i - 1]:
                    timer.logger.info("WorkInfo:" + str(kwargs))
                    timer.logger.info(str(i) + " Repaired")
                    timer.quick_repaired_cost += 1
                    timer.click(
                        BLOOD_BAR_POSITION[0][i][0],
                        BLOOD_BAR_POSITION[0][i][1],
                        delay=1.5,
                    )
            if switch_back:
                timer.click(200, 420, times=2, delay=1.5)  # 返回正常切换页面
    except AssertionError:
        raise ValueError(f"修理舰船的参数不合法, 请检查你的参数:{arg}")


def get_rewards(timer: Timer):
    """检查任务情况,如果可以领取奖励则领取"""
    timer.goto_game_page("main_page")
    if not timer.check_pixel((694, 457), bgr_color=(45, 89, 255)):
        return "no"
    timer.goto_game_page("mission_page")
    if timer.click_image(IMG.game_ui[15]):
        timer.relative_click(0.5, 0.5)
        timer.ConfirmOperation(must_confirm=1, timeout=5)
        return "ok"
    elif timer.click_image(IMG.game_ui[12]):
        timer.ConfirmOperation(must_confirm=1)
        return "ok"
    return "no"
    # timer.click(774, 502)


def RepairByBath(timer: Timer):
    """使用浴室修理修理时间最长的单位
    Args:
        timer (Timer): _description_
    """
    timer.goto_game_page("choose_repair_page")
    timer.click(115, 233)
    if not timer.identify_page("choose_repair_page"):
        if timer.identify_page("bath_page"):
            timer.set_page("bath_page")
        else:
            timer.set_page()


def SetAutoSupply(timer: Timer, type=1):
    timer.update_screen()
    NowType = int(timer.check_pixel((48, 508), (224, 135, 35)))
    if NowType != type:
        timer.click(44, 503, delay=0.33)


def Supply(timer: Timer, ship_ids=[1, 2, 3, 4, 5, 6], try_times=0):
    """补给指定舰船

    Args:
        timer (Timer): _description_
        ship_ids (list, optional): 补给舰船列表, 可以为单个整数表示单艘舰船. Defaults to [1, 2, 3, 4, 5, 6].

    Raises:
        ValueError: 补给失败
        TypeError: ship_ids 参数有误
    """
    if try_times > 3:
        raise ValueError("can't supply ship")

    if isinstance(ship_ids, int):
        ship_ids = [ship_ids]

    timer.click(293, 420)
    for x in ship_ids:
        if not isinstance(x, int):
            raise TypeError("ship must be represent as a int but get" + str(ship_ids))
        timer.click(110 * x, 241)

    if timer.is_bad_network(0):
        timer.process_bad_network("supply ships")
        Supply(timer, ship_ids, try_times + 1)


def ChangeShip(
    timer: Timer, fleet_id, ship_id=None, name=None, pre=None, ship_stats=None
):
    """切换舰船(不支持第一舰队)"""
    if fleet_id is not None:
        timer.goto_game_page("fight_prepare_page")
        MoveTeam(timer, fleet_id)
        if fleet_id >= 5:
            # 切换为预设编队
            # 暂不支持
            return

    # 切换单船
    if ship_stats is None:
        ship_stats = detect_ship_stats(timer)
    if name is None and ship_stats[ship_id] == -1:
        return
    timer.click(110 * ship_id, 250, delay=0)
    res = timer.wait_images(
        [IMG.choose_ship_image[1], IMG.choose_ship_image[2]], after_get_delay=0.4, gap=0
    )
    if res == 1:
        timer.click(839, 113)

    if name is None:
        timer.click(83, 167, delay=0)
        timer.wait_pages("fight_prepare_page", gap=0)
        return

    timer.click(700, 30, delay=0)
    # 等待输入框出现
    start_time = time.time()
    while not timer.check_pixel((955, 500), (255, 255, 255), screen_shot=True):
        time.sleep(0.2)
        if time.time() - start_time > 10:
            timer.log_screen(True)
            raise TimeoutError("等待输入框出现时超时")
    timer.logger.debug("输入框已经出现")
    # 输入名字检索
    timer.text(name)
    # 输入后随便点击获得检索结果
    timer.click(50, 50, delay=0.5)
    time.sleep(0.5)
    # OCR识别舰船
    if not name in timer.ship_names:
        timer.ship_names.append(name)
    ship_info = timer.recognize_ship(timer.get_screen()[:, :1048], timer.ship_names)

    # 查找识别结果中要选的舰船
    found_ship = next((ship for ship in ship_info if ship[1] == name), None)
    # 点击舰船
    if found_ship is None:
        timer.logger.error(f"Can't find ship {name}, ocr result:{ship_info}")
        if len(ship_info) == 0:
            timer.logger.warning("无法查找到任何舰船, 请检查舰船名是否有错误")
            raise ShipNotFoundErr()
        else:
            timer.logger.debug("Try to click the first ship")
            if ship_stats[ship_id] == -1:
                timer.click(83, 167, delay=0)
            else:
                timer.click(183, 167, delay=0)
    else:
        center = found_ship[0]
        rel_center = absolute_to_relative(center, timer.resolution)
        timer.relative_click(*rel_center)

    timer.wait_pages("fight_prepare_page", gap=0)


def ChangeShips(timer: Timer, fleet_id, ship_names):
    """更换编队舰船

    Args:
        fleet_id (int): 2~4,表示舰队编号

        ship_names (舰船名称列表):

    For example:
        ChangeShips(timer, 2, [None, "萤火虫", "伏尔塔", "吹雪", "明斯克", None, None])

    """
    timer.logger.info(f"Change Fleet {fleet_id} to {ship_names}")
    if fleet_id is not None:
        timer.goto_game_page("fight_prepare_page")
        MoveTeam(timer, fleet_id)
    if fleet_id == 1:
        raise ValueError("change member of fleet 1 is unsupported")
    ship_stats = detect_ship_stats(timer)
    for i in range(1, 7):
        if ship_stats[i] != -1:
            ChangeShip(timer, fleet_id, 1, None, ship_stats=ship_stats)
    ship_names = ship_names + [None] * 6
    for i in range(len(ship_names)):
        if ship_names[i] == "":
            ship_names[i] = None
    time.sleep(0.3)
    ship_stats = detect_ship_stats(timer)
    for i in range(1, 7):
        ChangeShip(timer, fleet_id, i, ship_names[i], ship_stats=ship_stats)


def get_new_things(timer: Timer, lock=0):
    pass


# 是否强制点击
def cook(timer: Timer, position: int, force_click=False):
    """食堂做菜
    Args:
        position (int): 第几个菜谱
        force_click (bool, optional): 当有菜正在生效时是否继续做菜. Defaults to False.
    """
    if position < 1 or position > 3:
        raise ValueError(f"不支持的菜谱编号:{position}")
    POSITION = [None, (318, 276), (420, 140), (556, 217)]
    timer.goto_game_page("canteen_page")
    timer.click(*POSITION[position])
    try:
        timer.click_image(IMG.restaurant_image.cook, timeout=7.5, must_click=True)
        if timer.image_exist(IMG.restaurant_image.have_cook):
            timer.logger.info("当前菜的效果正在生效")
            if force_click:
                timer.relative_click(0.414, 0.628)
                if timer.image_exist(IMG.restaurant_image.no_times):
                    timer.logger.info("今日用餐次数已经用尽")
                    timer.relative_click(0.788, 0.207)
                    return False
                timer.logger.info("做菜成功")
            else:
                timer.relative_click(0.65, 0.628)
                timer.logger.info("取消做菜")
                timer.relative_click(0.788, 0.207)
        return True

    except:
        timer.logger.warning(
            f"不支持的菜谱编号:{position}, 请检查该菜谱是否有效, 或者检查今日用餐次数是否已经用尽"
        )
        return False



import os
import platform
import re
import subprocess
import sys


def _fetch_requirements(path):
    with open(path, "r") as fd:
        return [r.strip() for r in fd.readlines()]


def _fetch_readme():
    with open("README.md", encoding="utf-8") as f:
        return f.read()


def find_version(filepath: str) -> str:
    """Extract version information from the given filepath.

    Adapted from https://github.com/ray-project/ray/blob/0b190ee1160eeca9796bc091e07eaebf4c85b511/python/setup.py
    """
    with open(filepath) as fp:
        version_match = re.search(
            r"^__version__ = ['\"]([^'\"]*)['\"]", fp.read(), re.M
        )
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")


def install_wheel():
    try:
        import wheel
    except ImportError:
        subprocess.run(["pip", "install", "wheel"])


install_wheel()


def _fetch_package_name():
    return "autowsgr"


from setuptools import find_namespace_packages, find_packages, setup
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel


# Custom wheel class to modify the wheel name
class bdist_wheel(_bdist_wheel):
    def finalize_options(self):
        _bdist_wheel.finalize_options(self)
        self.root_is_pure = False

    def get_tag(self):
        python_version = f"cp{sys.version_info.major}{sys.version_info.minor}"
        abi_tag = f"{python_version}"

        if platform.system() == "Linux":
            platform_tag = "manylinux1_x86_64"
        else:
            platform_tag = platform.system().lower()

        return python_version, abi_tag, platform_tag


# Setup configuration
setup(
    name=_fetch_package_name(),
    version=find_version("autowsgr/__init__.py"),
    packages=find_namespace_packages(
        include=["autowsgr*", "awsg*"],
        exclude=(
            "docs",
            "examples",
        ),
    ),
    include_package_data=True,
    description="Auto Warship Girls Framework.",
    long_description=_fetch_readme(),
    long_description_content_type="text/markdown",
    install_requires=_fetch_requirements("requirements.txt"),
    setup_requires=["wheel"],
    dependency_links=[
        "https://www.paddlepaddle.org.cn/packages/stable/cu123/",
        "https://download.pytorch.org/whl/cu123",
    ],
    python_requires=">=3.9",
    package_data={
        "": [
            "version.txt",
            "data/**",
            "requirements.txt",
            "bin/**",
            "c_src/**",
        ],  # 希望被打包的文件
    },
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Environment :: GPU :: NVIDIA CUDA",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: System :: Distributed Computing",
    ],
    cmdclass={"bdist_wheel": bdist_wheel},
)



import os
import subprocess
import time

from autowsgr.constants.custom_exceptions import ImageNotFoundErr
from autowsgr.constants.data_roots import MAP_ROOT, TUNNEL_ROOT
from autowsgr.constants.image_templates import IMG
from autowsgr.fight.battle import BattleInfo, BattlePlan
from autowsgr.fight.common import start_march
from autowsgr.game.game_operation import DestroyShip, get_ship, quick_repair
from autowsgr.game.get_game_info import detect_ship_stats
from autowsgr.ocr.ship_name import recognize
from autowsgr.port.ship import Fleet, count_ship
from autowsgr.timer import Timer
from autowsgr.utils.api_image import (
    crop_image,
    crop_rectangle_relative,
    cv2,
    cv_show_image,
)
from autowsgr.utils.io import count, yaml_to_dict

"""决战结构:
上层控制+单点战斗
"""


def is_ship(element):
    return element not in ["长跑训练", "肌肉记忆", "黑科技"]


def get_formation(fleet: Fleet, enemy: list):
    anti_sub = count(["CL", "DD", "CVL"], enemy)
    if fleet.exist("U-1206") and anti_sub <= 1:
        return 4
    elif anti_sub <= 0:
        return 4
    return 2


class DB:
    def __init__(self, *args, **kwargs):
        self.__dict__.update(kwargs)

    def make_decision(self, state):
        rule = self.__dict__
        if state in rule:
            return rule[state]


class DecisiveStats:
    def __init__(self, timer: Timer, chapter=6) -> None:
        # 选择战备舰队的点击位置
        self.timer = timer
        self.key_points = [
            [
                "",
            ]
        ]  # [chapter][map] (str)
        self.map_end = [
            "",
        ]  # [chapter] (str)
        self.enemy = [
            [
                [
                    "",
                ]
            ]
        ]  # [chapter][map][node(str)] (lst["", enemies])
        self.__dict__.update(
            yaml_to_dict(os.path.join(MAP_ROOT, "decisive_battle", f"3.yaml"))
        )
        self.score = 10
        self.level = 1  # 副官等级
        self.exp = 0
        self.need = 0  # 副官升级所需经验
        self.chapter = chapter  # 大关
        self.fleet = Fleet(self.timer)
        self.ships = set()
        self.ship_stats = [-1] * 7
        self.selections = []  # 获取战备舰队的元素

    def next(self):
        if self.node == self.map_end[self.chapter][self.map]:
            self.map += 1
            self.node = "A"
            return "quit" if (self.map == 4) else "next"
        self.node = chr(ord(self.node) + 1)
        return "continue"

    @property
    def enemy_now(self):
        return self.enemy[self.chapter][self.map][self.node]

    def reset(self):
        chapter = self.chapter
        self.__init__(self.timer)
        self.chapter = chapter

    def is_begin(self):
        return self.node == "A" and self.map == 1


class Logic:
    """决战逻辑模块"""

    def __init__(
        self,
        timer,
        stats: DecisiveStats,
        level1: list,
        level2: list,
        flagship_priority: list,
    ):
        self.timer = timer
        self.config = timer.config
        self.logger = timer.logger

        self.level1 = level1
        self.level2 = ["长跑训练", "肌肉记忆"] + self.level1 + level2 + ["黑科技"]
        self.flag_ships = flagship_priority
        self.stats = stats

    def _choose_ship(self, must=False):
        lim = 6
        score = self.stats.score
        if self.stats.fleet.count() <= 1:
            choose = self.level1
        elif self.stats.fleet.count() < 6:
            choose = [element for element in self.level2 if is_ship(element)]
        else:
            lim = score
            choose = self.level1 + [
                element for element in self.level2 if not is_ship(element)
            ]
        result = []
        for target in choose:
            if target in self.stats.selections.keys():
                cost = self.stats.selections[target][0]
                if score >= cost and cost <= lim:
                    score -= cost
                    result.append(target)
        if not result and must:
            result.append(list(self.stats.selections.keys())[0])
        return result

    def _use_skill(self):
        return 3 if (self.stats.node == "A") else 0

    def need_repair(self):
        return bool((1 in self.stats.ship_stats or 2 in self.stats.ship_stats))

    def _up_level(self):
        return bool((self.stats.need - self.stats.exp <= 5 and self.stats.score >= 5))

    def formation(self):
        pass

    def night(self):
        pass

    def get_best_fleet(self):
        ships = self.stats.ships
        self.logger.debug(f"拥有舰船: {ships}")
        best_ships = [
            "",
        ]
        for ship in self.level1:
            if ship not in ships or len(best_ships) == 7:
                continue
            best_ships.append(ship)
        for ship in self.level2:
            if ship not in ships or len(best_ships) == 7 or ship in self.level1:
                continue
            best_ships.append(ship)

        for flag_ship in self.flag_ships:
            if flag_ship not in best_ships:
                continue
            p = best_ships.index(flag_ship)
            best_ships[p], best_ships[1] = best_ships[1], best_ships[p]
            break

        for _ in range(len(best_ships), 7):
            best_ships.append("")
        self.logger.debug(f"当前最优：{best_ships}")
        return best_ships

    def _retreat(self):
        return bool((count_ship(self.get_best_fleet()) < 2))

    def _leave(self):
        return False


class DecisiveBattle:
    """决战控制模块
    目前仅支持 E5, E6, E4
    """

    def run_for_times(self, times=1):
        assert times >= 1
        self.start_fight()
        for _ in range(times - 1):
            self.reset_chapter()
            self.start_fight()

    def run(self):
        self.run_for_times()

    def __init__(
        self,
        timer: Timer,
        chapter=6,
        level1=["鲃鱼", "U-1206", "U-47", "射水鱼", "U-96", "U-1405"],
        level2=["U-81", "大青花鱼"],
        flagship_priority=["U-1405", "U-47", "U-96", "U-1206"],
        logic=None,
        repair_level=1,
        full_destroy=False,
        *args,
        **kwargs,
    ):
        """初始化控制模块
        Important Information:
            : 决战逻辑相当优秀, 不建议自行改写, 满配情况约需要 3~6 快速修复打完一轮 E6, 耗时约为 25mins
            : 请保证可以直接使用上次选船通关, 暂时不支持自动选船
        Args:
            timer (Timer): 记录器
            chapter (int, optional): 决战章节,请保证为 [1, 6] 中的整数. Defaults to 6.
            level1: 第一优先舰队, Defaults Recommend
            level2: 第二优先舰船, Defaults Recommend
            flagship_priority: 旗舰优先级, Defaults Recommend
            logic: 逻辑模块, unsupported since now
            repair_level: 维修策略，1为中破修，2为大破修，Defaults to 1.
            full_destroy: 船舱满了之后是否自动解装，默认不解装.
        Examples:
            : 若当前决战进度为还没打 6-2-A, 则应填写 chapter=6, map=1, node='A'
            : 可以支持断点开始
        """
        self.timer = timer
        self.config = timer.config
        self.repair_strategy = repair_level
        self.full_destroy = full_destroy
        assert chapter <= 6 and chapter >= 1
        self.stats = DecisiveStats(timer, chapter)
        if logic is None:
            self.logic = Logic(
                self.timer, self.stats, level1, level2, flagship_priority
            )
        self.__dict__.update(kwargs)

    def buy_ticket(self, use="steel", times=3):
        self.enter_decisive_battle()
        position = {"oil": 184, "ammo": 235, "steel": 279, "aluminum": 321}
        self.timer.click(458 * 0.75, 665 * 0.75, delay=1.5)
        self.timer.click(638, position[use], delay=1, times=times)
        self.timer.click(488, 405)

    def recognize_map(self):
        CHECK_POINT = {
            4: [(0.381, 0.436), (0.596, 0.636), (0.778, 0.521)],
            5: [(0.418, 0.378), (0.760, 0.477), (0.550, 0.750)],
            6: [(0.606, 0.375), (0.532, 0.703), (0.862, 0.644)],
        }
        check_point = CHECK_POINT[self.stats.chapter]
        for i, point in enumerate(check_point):
            if not self.timer.check_pixel(
                (int(point[0] * 960), int(point[1] * 540)), (250, 244, 253), 30, True
            ):
                self.timer.logger.info(f"识别决战地图参数, 第 {i} 小节正在进行")
                return i
        self.timer.logger.info(f"识别决战地图参数, 第 3 小节正在进行")
        return 3

    def recognize_node(self):
        position = self.timer.wait_images_position(
            IMG.fight_image[18:20], confidence=0.7
        )
        self.timer.logger.debug(f"Ship_ICON position: {position}")
        cropped_image = crop_rectangle_relative(
            self.timer.get_screen(), position[0] / 960 - 0.03, 0, 0.042, 1
        )
        image_path = os.path.join(TUNNEL_ROOT, "1.PNG")
        cv2.imwrite(image_path, cropped_image)
        processor = os.path.join(TUNNEL_ROOT, "process_recognize_map.exe")
        subprocess.run([processor, image_path])
        cropped_image = cv2.imread(image_path)
        with open(os.path.join(TUNNEL_ROOT, "1.out"), mode="r") as f:
            result = f.read()
            self.timer.logger.info(f"识别决战地图参数, 第 {result[0]} 节点正在进行")
            return result[0]
        # result = recognize(cropped_image, "ABCDEFGHIJK")
        # return result[0][1]

    def detect(self, type="enter_map"):
        """检查当前关卡状态
        Args:
            type:
                enter_map: 决战入口检查
                running: 检查地图是否在战斗准备页面

        Returns:
            str: ['challenging', 'refreshed', 'refresh']
            str: ['fight_prepare', 'map']
        """
        if type == "enter_map":
            _res = ["cant_fight", "challenging", "refreshed", "refresh"]
            res = self.timer.wait_images(
                IMG.decisive_battle_image[3:7], after_get_delay=0.2
            )
        if type == "running":
            _res = ["map", "fight_prepare"]
            res = self.timer.wait_images(
                [IMG.decisive_battle_image[1]]
                + IMG.identify_images["fight_prepare_page"],
                gap=0.03,
                after_get_delay=0.2,
            )
        return _res[res]

    def _go_map_page(self):
        if self.detect("running") == "fight_prepare":
            self.timer.click(30, 30)
            self.timer.wait_image(IMG.decisive_battle_image[1])

    def go_fleet_page(self):
        if self.detect("running") == "map":
            self.timer.click(900 * 0.75, 667 * 0.75)
            try:
                self.timer.wait_images(
                    IMG.identify_images["fight_prepare_page"],
                    timeout=5,
                    after_get_delay=1,
                )
            except:
                self.timer.logger.warning("进入出征准备页面失败，正在重试")
                self.go_fleet_page()

    def repair(self):
        self.go_fleet_page()
        quick_repair(
            self.timer, self.repair_strategy
        )  # TODO：我的中破比很高，先改成只修大破控制一下用桶
        # quick_repair(self.timer, 2)

    def next(self):
        res = self.stats.next()
        if res in ["next", "quit"]:
            self.timer.ConfirmOperation(timeout=5, must_confirm=1)  # 确认通关
            self.timer.ConfirmOperation(timeout=5, must_confirm=1)  # 确认领取奖励
            get_ship(self.timer)
        return res

    def choose(self, refreshed=False, rec_only=False):
        # ===================获取备选项信息======================
        # 右上角资源位置
        RESOURCE_AREA = ((0.911, 0.082), (0.974, 0.037))
        # 船名位置
        SHIP_X = [
            (0.195, 0.305),
            (0.318, 0.429),
            (0.445, 0.555),
            (0.571, 0.677),
            (0.695, 0.805),
        ]
        SHIP_Y = (0.715, 0.685)
        # 船的费用
        COST_AREA = ((0.195, 0.808), (0.805, 0.764))
        # 选择位置
        CHOOSE_X = [0.25, 0.375, 0.5, 0.625, 0.75]
        CHOOSE_Y = 0.5

        screen = self.timer.get_screen()
        try:
            self.stats.score = self.timer.recognize_number(
                crop_image(screen, *RESOURCE_AREA),
                # rgb_select=(250, 250, 250),
                # tolerance = 50,
            )[1]
        except:
            # TODO: 提高OCR对单个数字的识别率
            self.timer.logger.warning("读取当前可用费用失败")
            self.stats.score = 0
        self.timer.logger.debug(f"当前可用费用为：{self.stats.score}")
        results = self.timer.recognize_number(
            crop_image(screen, *COST_AREA),
            extra_chars="x",
            multiple=True,
        )
        costs = [t[1] for t in results]
        _costs, ships, real_position = [], [], []
        for i, cost in enumerate(costs):
            try:
                if cost > self.stats.score:
                    continue
            except Exception as e:
                self.timer.logger.warning(f"读取购买费用出错，错误如下:\n {e}")
                continue
            ships.append(
                self.timer.recognize(
                    crop_image(
                        screen, (SHIP_X[i][0], SHIP_Y[0]), (SHIP_X[i][1], SHIP_Y[1])
                    ),
                    candidates=self.timer.ship_names,
                )[1]
            )
            _costs.append(cost)
            real_position.append(i)
        # print("Scan result:", costs)
        costs = _costs
        selections = {
            ships[i]: (costs[i], (CHOOSE_X[real_position[i]], CHOOSE_Y))
            for i in range(len(costs))
        }
        if rec_only:
            return
        # ==================做出决策===================
        self.stats.selections = selections
        self.timer.logger.debug("可购买舰船：", selections)
        choose = self.logic._choose_ship(
            must=(self.stats.map == 1 and self.stats.node == "A" and refreshed == True)
        )
        if len(choose) == 0 and refreshed == False:
            self.timer.click(380, 500)  # 刷新备选舰船
            return self.choose(True)

        for target in choose:
            cost, p = selections[target]
            self.stats.score -= cost
            self.timer.logger.debug(f"选择购买：{target}，花费：{cost}，点击位置：{p}")
            self.timer.relative_click(*p)
            if is_ship(target):
                self.stats.ships.add(target)
        self.timer.click(580, 500)  # 关闭/确定

    def up_level_assistant(self):
        self.timer.click(75, 667 * 0.75)
        self.stats.score -= 5

    def use_skill(self, type=3):
        SKILL_POS = (0.2143, 0.894)
        SHIP_AREA = ((0.26, 0.715), (0.74, 0.685))

        self.timer.relative_click(*SKILL_POS)
        if type == 3:
            ship_results = self.timer.recognize(
                crop_image(self.timer.get_screen(), *SHIP_AREA),
                candidates=self.timer.ship_names,
                multiple=True,
            )
            ships = [ship[1] for ship in ship_results]
            self.timer.logger.info(f"使用技能获得: {ships}")
            for ship in ships:
                self.stats.ships.add(ship)
        self.timer.relative_click(*SKILL_POS, times=2, delay=0.3)

    def leave(self):
        self.timer.click(36, 33)
        self.timer.click(360, 300)

    def _get_chapter(self):
        CHAPTER_AREA = ((0.818, 0.867), (0.871, 0.826))
        text = self.timer.recognize(
            crop_image(self.timer.get_screen(), *CHAPTER_AREA),
            allowlist="Ex-0123456789",
            rgb_select=(247, 221, 82),
            tolerance=50,
        )[1]
        return int(text[-1])

    def _move_chapter(self):
        if self._get_chapter() < self.stats.chapter:
            self.timer.click(900, 507)
        elif self._get_chapter() > self.stats.chapter:
            self.timer.click(788, 507)
        else:
            return
        self._move_chapter()

    def enter_decisive_battle(self):
        self.timer.goto_game_page("decisive_battle_entrance")
        self.timer.click(115, 113, delay=1.5)
        self.detect()

    def enter_map(self, check_map=True):
        self.stats.node = "A"
        if check_map:
            self.enter_decisive_battle()
            self._move_chapter()
            stats = self.detect()
            self.stats.map = self.recognize_map()
            if stats == "refresh":
                entered = self.reset_chapter()
                if entered:
                    return
                stats = "refreshed"
            if stats == "refreshed":
                # 选用上一次的舰船并进入
                if self.check_dock_full():
                    return "full_destroy_success"
                self.timer.click(500, 500, delay=0.25)
                self.stats.map = 1
                stats = self.detect()
                if stats == "cant_fight":
                    raise SystemError("其他关卡正在进行挑战")
                for i in range(5):
                    self.timer.click_image(
                        IMG.decisive_battle_image[7], timeout=12, must_click=True
                    )
                    self.timer.click(873, 500)
                    if (
                        self.timer.wait_images(
                            [
                                IMG.decisive_battle_image[1],
                                IMG.decisive_battle_image[3],
                            ],
                            timeout=10,
                            gap=0.03,
                        )
                        is not None
                    ):
                        break
                if i > 3:
                    raise TimeoutError("选择决战舰船失败")
            else:
                self.timer.click(500, 500, delay=0)
        else:
            self.detect()
            self.stats.map = self.recognize_map()
            self.timer.click(500, 500, delay=0)

        if self.check_dock_full():
            return "full_destroy_success"

        res = self.timer.wait_images(
            [IMG.decisive_battle_image[1], IMG.decisive_battle_image[3]],
            timeout=10,
            gap=0.03,
        )
        if res is None:
            raise ImageNotFoundErr("Can't Identify on enter_map")
        return "other chapter is running" if (res == 1) else "ok"

    def check_dock_full(self):
        """
        检查船舱是否满，船舱满了自动解装
        """
        if self.timer.wait_images(IMG.symbol_image[12], timeout=2) is not None:
            if self.full_destroy:
                self.timer.relative_click(0.38, 0.565)
                DestroyShip(self.timer)
                self.enter_map()
                return True
            else:
                self.timer.logger.warning("船舱已满, 但不允许解装")
                return False
        return False

    def retreat(self):
        self._go_map_page()
        self.timer.click(36, 33)
        self.timer.click(600, 300)

    def _get_exp(self, retry=0):
        EXP_AREA = ((0.018, 0.854), (0.092, 0.822))
        try:
            self.stats.exp = 0
            self.stats.need = 20
            try:
                src = self.timer.recognize(
                    crop_image(self.timer.get_screen(), *EXP_AREA),
                    allowlist="Lv.(/)0123456789",
                )[1]
                try:
                    index1 = src.index("(")
                except ValueError:
                    index1 = float("inf")  # 如果没有找到，设置为无穷大

                try:
                    index2 = src.index("（")
                except ValueError:
                    index2 = float("inf")  # 如果没有找到，设置为无穷大
                i1 = min(index1, index2)
                if i1 == float("inf"):
                    raise ValueError("未找到 '(' 或 '（'")
                i2 = src.index("/")
                src = src.rstrip(")）")
                self.stats.exp = int(src[i1 + 1 : i2])
                self.stats.need = int(src[i2 + 1 :])
                self.timer.logger.debug(
                    f"当前经验：{self.stats.exp}，升级需要经验：{self.stats.need}"
                )
            except:
                self.timer.logger.warning("识别副官升级经验数值失败")
        except:
            if retry > 1:
                self.timer.logger.warning("重新读取 exp 失败, 退出逻辑")
                raise BaseException()  # ToDo: 定义对应的 Exception

            self.timer.logger.warning("读取exp失败，五秒后重试")
            self.timer.click(580, 500)
            time.sleep(1)
            self._get_exp(retry + 1)

    def _before_fight(self):
        self.timer.logger.info("Test")
        if self.timer.wait_image(IMG.confirm_image[1:], timeout=1) != False:
            self.timer.click(300, 225)  # 选上中下路
            self.timer.ConfirmOperation(must_confirm=1)
        if self.timer.wait_image(
            [IMG.decisive_battle_image[2], IMG.decisive_battle_image[8]], timeout=2
        ):
            self.choose()  # 获取战备舰队
        self._get_exp()
        self.stats.node = self.recognize_node()
        # 升级副官, 现在这功能坏掉了
        # while self.logic._up_level():
        #     self.up_level_assistant()
        #     self._get_exp()
        if self.logic._use_skill():
            self.use_skill(self.logic._use_skill())

        if self.stats.fleet.empty() and not self.stats.is_begin():
            self._check_fleet()
        _fleet = self.logic.get_best_fleet()
        if self.logic._retreat():
            self.retreat()
            return "retreat"
        if self.logic._leave():
            self.leave()
            return "leave"
        if self.stats.fleet != _fleet:
            self._change_fleet(_fleet)
            self.stats.ship_stats = detect_ship_stats(self.timer)
        if self.logic.need_repair():
            self.repair()

    def _after_fight(self):
        self.timer.logger.info(self.stats.ship_stats)

    def _check_fleet(self):
        self.stats.ships.clear()
        self.go_fleet_page()
        self.stats.fleet.detect()
        for ship in self.stats.fleet.ships[1:]:
            self.stats.ships.add(ship)
        self.stats.ship_stats = detect_ship_stats(self.timer)
        self.timer.relative_click(0.1, 0.5)
        res = self.timer.wait_images(
            IMG.choose_ship_image[1:3] + [IMG.choose_ship_image[4]],
            after_get_delay=0.4,
            gap=0,
            timeout=16,
        )
        if res == None:
            raise TimeoutError("选择舰船时点击超时")
        ships = self.timer.recognize_ship(
            self.timer.get_screen()[:, :1048], self.timer.ship_names
        )
        self.timer.logger.info(f"其它可用舰船：{[item[1] for item in ships]}")
        for ship in ships:
            self.stats.ships.add(ship[1])
        self.timer.relative_click(0.05, 0.05)
        res = self.timer.wait_images(
            IMG.identify_images["fight_prepare_page"],
            gap=0.03,
            after_get_delay=0.2,
        )

    def _during_fight(self):
        formation = get_formation(self.stats.fleet, self.stats.enemy_now)
        night = (
            self.stats.node in self.stats.key_points[self.stats.chapter][self.stats.map]
        )
        plan = DecisiveBattlePlan(
            self.timer,
            formation=formation,
            night=night,
            ship_stats=self.stats.ship_stats,
        )
        plan.run()
        self.stats.ship_stats = plan.Info.fight_history.get_fight_results()[
            -1
        ].ship_stats

    def _change_fleet(self, fleet):
        self.go_fleet_page()
        self.stats.fleet.set_ship(fleet, order=True, search_method=None)

    def fight(self):
        # try:
        res = self._before_fight()
        # except BaseException as e:
        #     print(e)
        #     if self.stats.map == 1 and self.stats.node == 'A':
        #         # 处理临时 BUG (https://nga.178.com/read.php?tid=34341326)
        #         print(e, "Temporary Game BUG, Processing...")
        #         self.timer.restart()
        #         self.enter_map()
        #         self._reset()
        #         return 'continue'

        if res == "retreat":
            self.enter_map(check_map=False)
            self._reset()
            return self.fight()
        self._during_fight()
        self._after_fight()
        return self.next()

    def start_fight(self):
        self.enter_map()
        while True:
            res = self.fight()
            if res == "quit":
                return
            elif res == "next":
                self.enter_map(False)

    def reset_chapter(self):
        """使用磁盘重置关卡, 并重置状态"""
        # Todo: 缺少磁盘报错
        self._reset()
        self._move_chapter()
        self.timer.relative_click(0.5, 0.925)
        if self.check_dock_full():
            return True
        else:
            self.timer.ConfirmOperation(must_confirm=True)

    def _reset(self):
        self.stats.reset()


class DecisiveBattlePlan(BattlePlan):
    def __init__(self, timer: Timer, formation, night, ship_stats):
        super().__init__(timer, None)
        self.Info = DecisiveBattleInfo(timer)
        self.Info.ship_stats = ship_stats
        self.node.formation = formation
        self.node.night = night

    def _enter_fight(self, *args, **kwargs):
        return start_march(self.timer)


class DecisiveBattleInfo(BattleInfo):
    def __init__(self, timer: Timer):
        super().__init__(timer)
        self.end_page = "unknown_page"
        self.state2image["battle_page"] = [IMG.decisive_battle_image[1], 5]
        self.state2image["result"] = [IMG.fight_image[3], 60, 0.7]
